// **************************************************************************
//
//	Queuer - Recovery - Banned - Guru
//
// **************************************************************************
#include "include.h"

// **************************************************************************
// Queuer
// **************************************************************************

/*
 The queuer is the part that will trigger
 the callback when required. For posting replies
 and contact point change notifications.
 Once a notification has been sent, the entry
 is removed from the queuer database, independant
 if the other peer will actually come to get
 the reply.

 The queuer has entries per contact point.
 It is a seperate thread that mostly sleeps.
*/

enum Qaction {
	REPLY_NOTIFY, PCR_NOTIFY, PCP_NOTIFY, APING
};

class Qentry : public dlistNodeAuto
{
   public:
	Qentry (dlist*, Qaction, unsigned int, cryptotype);
	Qaction Action;
	unsigned int ID;
	cryptotype secret;
};

class Qpoint : public dbsNodeCaseStr
{
   public:
	Qpoint ();
	dlist Entries;
	void addentry (Qaction, unsigned int, cryptotype = 0);
	int Notify ();
	void PointChange (char*);
	~Qpoint ();
};

Qentry::Qentry (dlist *D, Qaction A, unsigned int id, cryptotype s)
	: dlistNodeAuto (D)
{
	Action = A;
	ID = id;
	secret = s;
}

Qpoint::Qpoint () : dbsNodeCaseStr (&Globals.QueuerTree)
{
	Name = StrDup (DBS_STRQUERY);
}

void Qpoint::addentry (Qaction A, unsigned int ID, cryptotype s)
{
	Qentry *e = (Qentry*) Entries.Start;

	if (e && A == APING) return;

	for (; e; e = (Qentry*) Entries.Next (e))
		if (e->Action == A && e->ID == ID)
			if (A != PCP_NOTIFY) break;
			else if (s == e->secret) break;

	if (!e) new Qentry (&Entries, A, ID, s);
}

static post *keyPost (unsigned int k)
{
	return (post*) Globals.Posts.intFind (k);
}

/*
  Returns:
	0	: contact point not reachable
	1	: Ok, one notification message sent (oldest)
	2	: No more messages for this point left (destruct)
*/
int Qpoint::Notify ()
{
#define QE ((Qentry*)Entries.Start)
	i_argstream I;
	char author [MAX_AUTHOR], oas [200];
	int version;
	post *p;
	o_argstream O (oas, sizeof oas);
	unsigned int i;

DPRINTF (DB_SLOT3, "$$$ Queuer->Notify for [%s] Entries.cnt=%i\n", Name, Entries.cnt);
	if (Entries.cnt == 0) return 2;

	if (!contact (Name, I, author, version)) return 0;

	for (i = 0; i < Globals.nPS; i++)
		if (Globals.PingRequested [i] == this) {
			if (Globals.nPS + 1 < Q_MAX_PING)
				Globals.PingSuccess [Globals.nPS++] =
					StrDup (Name);
			for (--Globals.nPR; i < Globals.nPR; i++)
				Globals.PingRequested [i] =
				Globals.PingRequested [i - 1];
			break;
		}

	sem_wait (&Globals.DatabaseLock);
	switch (QE->Action) {
	case REPLY_NOTIFY:
		/* PROTOCOL-STUFF-LOCATION */
		O.putarg (Globals.PMSGtext [HAVE_REPLY]);
		O.putMsgID (QE->ID, keyPost (QE->ID)->rkey);
		O.putarg (Globals.ThisPoint);
		break;
	case PCR_NOTIFY:	// other has parent we have reply
		if ((p = keyPost (QE->ID)) && p->zombie) {
			O.putarg (Globals.PMSGtext [POINT_CHANGE_REPLY]);
			O.putarg (MsgIDofPRL (p->parent));
			O.putarg (p->secret);
			O.putarg (Globals.ThisPoint);
		}
		break;
	case PCP_NOTIFY:	// other has reply we have parent
		if ((p = keyPost (QE->ID)) && p->needsupdate (QE->secret)) {
			O.putarg (Globals.PMSGtext[POINT_CHANGE_PARENT]);
			O.putarg (p->msgidsec (QE->secret));
			O.putarg (QE->secret);
			O.putarg (Globals.ThisPoint);
		}
		break;
	case APING:
		O.putarg (Globals.PMSGtext [PING]);
		break;
	}
	sem_post (&Globals.DatabaseLock);

	Entries.dremove (Entries.Start);

	O.writeout (I.fd);

	return 1;
}

void Qpoint::PointChange (char *np)
{
	DBS_STRQUERY = np;
	if (!Globals.QueuerTree.dbsFind ()) {
		Globals.QueuerTree.dbsRemove (this);
		addself (&Globals.QueuerTree);
		delete [] Name;
		Name = StrDup (np);
	}
}

Qpoint::~Qpoint ()
{
	Globals.QueuerTree.dbsRemove (this);
	delete [] Name;
}

// //////////////////////////////////////////////////////////////////////

class do_Qpoint : public dbsVTask
{
   public:
	Qpoint **qpad;
	int qpi;
	do_Qpoint () { qpi = 0; }
	void Do (dbsNode*);
};

void do_Qpoint::Do (dbsNode *d)
{
	int r;
	while ((r = ((Qpoint*)d)->Notify ()) == 1);
	if (r == 2) qpad [qpi++] = (Qpoint*) d;
}

static void QAwakening ()
{
	int i;
	do_Qpoint A;
	A.qpad = (Qpoint**) alloca (sizeof (Qpoint*) * Globals.QueuerTree.nnodes);

	sem_wait (&Globals.QueuerLock);
	Globals.QueuerTree.copytree (&A);

	for (i = 0; i < A.qpi; i++)
		delete A.qpad [i];
	sem_post (&Globals.QueuerLock);
}

void QAdapt (char *op, char *np)
{
	Qpoint *p;

	sem_wait (&Globals.QueuerLock);
	DBS_STRQUERY = op;
	if ((p = (Qpoint*) Globals.QueuerTree.dbsFind ()))
		p->PointChange (np);
	sem_post (&Globals.QueuerLock);
}

void QNewReply (char *to, unsigned int key)
{
	Qpoint *p;

DPRINTF (DB_SLOT3, "# # # #QNreply to [%s]\n", to);
	sem_wait (&Globals.QueuerLock);
	DBS_STRQUERY = to;
	if (!(p = (Qpoint*) Globals.QueuerTree.dbsFind ()))
		p = new Qpoint ();
	p->addentry (REPLY_NOTIFY, key);
	sem_post (&Globals.QueuerLock);
}

void QPing (char *peer)
{
	Qpoint *p;
	unsigned int i;

	sem_wait (&Globals.QueuerLock);
	if (Globals.nPR + 1 < Q_MAX_PING) {
		DBS_STRQUERY = peer;
		if (!(p = (Qpoint*) Globals.QueuerTree.dbsFind ()))
			p = new Qpoint ();
		for (i = 0; i < Globals.nPR; i++)
			if (Globals.PingRequested [i] == (dbsNode*) p)
				break;
		if (i == Globals.nPR) {
			p->addentry (APING, 0);
			Globals.PingRequested [Globals.nPR++] = (dbsNode*) p;
		}
	}
	sem_post (&Globals.QueuerLock);
}

void QChPoint (char *to, unsigned int key, cryptotype s)
{
	Qpoint *p;

	DBS_STRQUERY = to;
	if (!(p = (Qpoint*) Globals.QueuerTree.dbsFind ()))
		p = new Qpoint ();
	p->addentry ((s) ? PCP_NOTIFY : PCR_NOTIFY, key, s);
}

static void *QMain (void*)
{
	// the server must post the semaphore when its ready
	// to accept clients *******************************
	sem_wait (&Globals.DatabaseLock);

	while (!Globals.Shutdown) {
		QAwakening ();
		sleep (Globals.QueuerSleep);
	}

	return NULL;
}

void QStart ()
{
	pthread_t t;

	pthread_create (&t, NULL, QMain, NULL);
	pthread_detach (t);
}

/* HCC RESOLVE:
pthread_create()
	QMain (Pv)
END */
// **************************************************************************
// Recovery
// **************************************************************************

/*
 The partial information is very critical.
 If any partial information is lost this affects
 other posters (in relation to our posts); it can still
 be fixed with guru human-to-human arrangements.

 Of course, just dumping each post information is safe
 enough. The recovery file is a diff of the post
 information on the disk and the changes. So the recovery
 procedure acts like buffering of the safety stuff.

 New posts do not belong in the modification activities
 and are dumped immediately.
*/

/*
 save only modified posts before resetting the recovery
*/

static void ResetRecovery ()
{
	if (Globals.RecoveryFd)
		close (Globals.RecoveryFd);
	unlink (RECOVERY_FILE);
	Globals.RecoveryFd = open (RECOVERY_FILE, O_WRONLY | O_CREAT, S_IRWXU);
}

static void dump_mods ()
{
	unsigned int i;
	post *p;

	for (i = 0; i < Globals.RDiff.c; i++)
		if ((p = (post*) Globals.Posts.intFind
			(Globals.RDiff.Modified [i])))
			p->dump_ipi ();
	Globals.RDiff.c = 0;
	Globals.RDiff.last = 0;

	ResetRecovery ();
}

static void modified (unsigned int id)
{
	unsigned int i;

	for (i = 0; i < Globals.RDiff.c; i++)
		if (Globals.RDiff.Modified [i] == id) return;

	Globals.RDiff.Modified [Globals.RDiff.c++] = id;

	if (Globals.RDiff.c == R_MAX_MODIFIED)
		dump_mods ();
	else if (Globals.RDiff.last && time (NULL) - Globals.RDiff.last > R_MAX_TIME)
		dump_mods ();
	else Globals.RDiff.last = time (NULL);
}

//
//	runtime recovery calls
//

enum ract {
	NR, DR, PO, UP, UR
};

void RNewRep (unsigned int ID, partial_post *pp)
{
	unsigned int f, sp, st, sa;
	char *dd;

	f = sizeof (ract) +sizeof ID +sizeof pp->date
		+sizeof pp->secret +sizeof pp->crc;
	sp = strlen (pp->point) +1;
	st = strlen (pp->title) +1;
	sa = strlen (pp->author) +1;

	dd = (char*) alloca (f + sp + st + sa);

	*((ract*) dd) = NR;
	*((unsigned int*) (dd + sizeof (ract))) = ID;
	*((time_t*) (dd + sizeof (ract) + sizeof ID)) = pp->date;
	*((cryptotype*) (dd + sizeof (ract) + sizeof ID
		 + sizeof pp->date)) = pp->secret;
	*((unsigned long*) (dd + sizeof (ract) + sizeof ID
		 + sizeof pp->date + sizeof pp->secret)) = pp->crc;
	strcpy (dd + f, pp->point);
	strcpy (dd + f + sp, pp->title);
	strcpy (dd + f + st, pp->author);
	write (Globals.RecoveryFd, dd, f + sp + st + sa);

	modified (ID);
}

void RDelRep (unsigned int ID, char *point)
{
	unsigned int l = sizeof (ract) + strlen (point) + 1;
	char *dd = (char*) alloca (l);

	*((ract*) dd) = DR;
	strcpy (dd + sizeof (ract), point);
	write (Globals.RecoveryFd, dd, l);

	modified (ID);
}

void Rposted (unsigned int ID, cryptotype secret)
{
	char dd [sizeof (ract) + sizeof ID + sizeof secret];

	*((ract*) dd) = PO;
	*((unsigned int*) (dd + sizeof (ract))) = ID;
	*((cryptotype*) (dd + sizeof (ract) + sizeof ID)) = secret;
	write (Globals.RecoveryFd, &dd [0], sizeof dd);

	modified (ID);
}

void Rupdate_p (unsigned int ID, cryptotype secret)
{
	char dd [sizeof (ract) + sizeof ID + sizeof secret];

	*((ract*) dd) = UP;
	*((unsigned int*) (dd + sizeof (ract))) = ID;
	*((cryptotype*) (dd + sizeof (ract) + sizeof ID)) = secret;
	write (Globals.RecoveryFd, &dd [0], sizeof dd);

	modified (ID);
}

void Rupdate_r (unsigned int ID, cryptotype os, cryptotype ns)
{
	char dd [sizeof (ract) + sizeof ID + 2 * sizeof (cryptotype)];

	*((ract*) dd) = UR;
	*((unsigned int*) (dd + sizeof (ract))) = ID;
	*((cryptotype*) (dd + sizeof (ract) + sizeof ID)) = os;
	*((cryptotype*) (dd + sizeof (ract) + sizeof ID + sizeof os)) = ns;
	write (Globals.RecoveryFd, &dd [0], sizeof dd);

	modified (ID);
}

// ####################################################################
//			recovery in action
// ####################################################################

static bool fgetstr (char *r, int max, FILE *F)
{
	int i = 0, x;

	while (i < max) {
		if ((x = fgetc (F)) == EOF) return false;
		if ((r [i++] = x) == 0)
			return true;
	}
	return false;
}	

static void RECOVER (FILE *F)
{
	unsigned int rc;
	ract a;
	cryptotype s1, s2;
	unsigned int ID;
	char prl [MAX_PRL], title [MAX_TITLE], author [MAX_AUTHOR];
	time_t dt;
	unsigned long cr;
	post *p;

	for (rc = 0; fread (&a, sizeof a, 1, F) == 1; rc++)
	switch (a) {
	case NR:
		if (fread (&ID, sizeof ID, 1, F) != 1) goto broken;
		if (fread (&dt, sizeof dt, 1, F) != 1) goto broken;
		if (fread (&s1, sizeof s1, 1, F) != 1) goto broken;
		if (fread (&cr, sizeof cr, 1, F) != 1) goto broken;
		if (!fgetstr (prl, MAX_PRL, F)) goto broken;
		if (!fgetstr (title, MAX_TITLE, F)) goto broken;
		if (!fgetstr (author, MAX_AUTHOR, F)) goto broken;
		if ((p = (post*) Globals.Posts.intFind (ID))) {
			if (p->canpost (prl, NULL))
			p->addreply (StrDup (prl), StrDup (title),
				     StrDup (author), dt, cr, s1);
		}
		break;
	case DR:
		if (fread (&ID, sizeof ID, 1, F) != 1) goto broken;
		if (!fgetstr (prl, MAX_PRL, F)) goto broken;
		if ((p = (post*) Globals.Posts.intFind (ID))) {
			p->rmvreply (prl);
		}
		break;
	case PO:
		if (fread (&ID, sizeof ID, 1, F) != 1) goto broken;
		if (fread (&s1, sizeof s1, 1, F) != 1) goto broken;
		if ((p = (post*) Globals.Posts.intFind (ID)))
			p->secret = s1;
		break;
	case UP:
		if (fread (&ID, sizeof ID, 1, F) != 1) goto broken;
		if (fread (&s1, sizeof s1, 1, F) != 1) goto broken;
		if ((p = (post*) Globals.Posts.intFind (ID))) {
			p->secret = s1;
			p->zombie = false;
		}
		break;
	case UR:
		if (fread (&ID, sizeof ID, 1, F) != 1) goto broken;
		if (fread (&s1, sizeof s1, 1, F) != 1) goto broken;
		if (fread (&s2, sizeof s2, 1, F) != 1) goto broken;
		if ((p = (post*) Globals.Posts.intFind (ID)))
			p->updated (s1, s2);
		break;
	default:
		goto broken;
	}

	goto breakbreak;

broken:
	fputs ("Recovery file broken..\n", stderr);

breakbreak:
	fprintf (stderr, "%i recovery actions. Ok\n", rc);
	fclose (F);
}

void PrepareRecovery ()
{
	if (access (RECOVERY_FILE, F_OK) == 0
	&& filesize (RECOVERY_FILE) > 0) {
		fprintf (stderr, "Recovery file found!\n");
		RECOVER (fopen (RECOVERY_FILE, "r"));
		SaveIPI ();
	}
	ResetRecovery ();
}

void CloseRecovery ()
{
	dump_mods ();
}

// **************************************************************************
// Banned :(
// **************************************************************************

class saveBan : public dbsVTask
{
   public:
	FILE *banf;
	saveBan (FILE *f) : banf (f) {}
	void Do (dbsNode*);
};

void saveBan::Do (dbsNode *d)
{
	fputs (((dbsNodeCaseStr*)d)->Name, banf);
	fputc ('\n', banf);
}

void SaveBanned ()
{
	saveBan A (fopen (BAN_FILE, "w"));

	Globals.Banned.copytree (&A);
	fclose (A.banf);
}

void LoadBanned ()
{
	FILE *f;
	char e [MAX_PRL];

	if ((f = fopen (BAN_FILE, "r"))) {
		while (fgets (e, MAX_PRL, f))
		if (*e != 0) {
			e [strlen (e) - 1] = 0;
			DBS_STRQUERY = e;
			Globals.Banned.dbsFind ();
			new dbsNodeCaseStr (&Globals.Banned, StrDup (e));
		}
		fclose (f);
	}
}

void Ban (char *bad)
{
	DBS_STRQUERY = bad;
	if (!Globals.Banned.dbsFind ())
		new dbsNodeCaseStr (&Globals.Banned, StrDup (bad));
}

bool Banned (char *cp)
{
	DBS_STRQUERY = cp;
	if (Globals.Banned.dbsFind ()) return true;
	if (!strncasecmp (cp, "/loc/", 5) && slashcount (cp) == 4) {
		char bloc [MAX_PRL];
		strcpy (bloc, cp);
		*strrchr (bloc, '/') = 0;
		DBS_STRQUERY = bloc;
		if (Globals.Banned.dbsFind ()) return true;
	}

	return false;
}

// **************************************************************************
// GURU Meditation
// **************************************************************************

class re_move_user : public intVTask
{
	char *OP, *NP;
	int OPlen;
   public:
	re_move_user (char*, char*);
	void Do (intNode*);
};

class pattern_ch_point : public intVTask
{
	char *OP, *NP;
	int OPlen;
   public:
	pattern_ch_point (char*, char*);
	void Do (intNode*);
};

re_move_user::re_move_user (char *op, char *np)
{
	OPlen = strlen (OP = op);
	NP = np;
}

void re_move_user::Do (intNode *n)
{
	post *p = (post*) n;
	partial_post *pp = (partial_post*) p->replies.Start;

	if (!strncasecmp (p->parent, OP, OPlen))
		p->pcpchange (p->secret, NP, p->secret);

	for (; pp; pp = (partial_post*) p->replies.Next (pp))
		if (!strncasecmp (pp->point, OP, OPlen))
			p->pcrchange (pp->secret, NP, pp->secret);
}

pattern_ch_point::pattern_ch_point (char *op, char *np)
{
	OPlen = strlen (OP = op);
	NP = np;
}

void pattern_ch_point::Do (intNode *n)
{
	post *p = (post*) n;
	partial_post *pp = (partial_post*) p->replies.Start;
	char tmp [3 * MAX_PRL];

	if (!strncasecmp (p->parent, OP, OPlen))
		if (strlen (strcat (strcpy (tmp, NP), p->parent + OPlen))
			< MAX_PRL) {
			delete [] p->parent;
			p->parent = StrDup (tmp);
		}

	for (; pp; pp = (partial_post*) p->replies.Next (pp))
		if (!strncasecmp (pp->point, OP, OPlen))
			if (strlen (strcat (strcpy (tmp, NP),
			     pp->point + OPlen)) < MAX_PRL) {
				delete [] pp->point;
				pp->point = StrDup (tmp);
			}
}

void Guru_explict_cp (char *op, char *np)
{
	if (np [0] == 0 || np [strlen (np) - 1] == '/') return;
	re_move_user A (op, np);
	sem_wait (&Globals.DatabaseLock);
	Globals.Posts.foreach (&A);
	sem_post (&Globals.DatabaseLock);
}

void Guru_pattern_cp (char *op, char *np)
{
	pattern_ch_point A (op, np);
	sem_wait (&Globals.DatabaseLock);
	Globals.Posts.foreach (&A);
	sem_post (&Globals.DatabaseLock);
}
